import heapq

from manim import *

from manim_dsa.m_collection.m_array import *
from manim_dsa.m_collection.m_stack import *
from manim_dsa.m_graph.m_graph import *
from manim_dsa.m_variable.m_variable import *


class IterativeDfs(Scene):
    def dfs(self, graph, start):
        mGraph = (
            MGraph(graph, style=MGraphStyle.PURPLE)
            .scale(0.7)
            .node_layout()
            .to_edge(LEFT)
            .shift(DR)
        )
        mStack = MStack(style=MStackStyle.BLUE).scale(0.7).to_edge(RIGHT).shift(DL)
        self.play(Create(mGraph))
        self.play(Create(mStack))
        visited = {}
        stack = [start]
        prevList = [None]
        self.play(mStack.animate.append(start))
        for node in graph:
            visited[node] = False
        while stack:
            node = stack.pop()
            self.play(mStack.animate.pop())
            prev = prevList.pop()
            if prev and not visited[node]:
                self.play(mGraph[(prev, node)].animate.highlight())
            if not visited[node]:
                self.play(mGraph[node].animate.highlight())
            visited[node] = True
            for neighbor in graph[node]:
                if not visited[neighbor]:
                    stack.append(neighbor)
                    self.play(mStack.animate.append(neighbor))
                    prevList.append(node)

    def construct(self):
        graph = {
            "0": ["1", "2"],
            "1": ["0", "2", "3", "4"],
            "2": ["0", "1"],
            "3": ["1", "5"],
            "4": ["1"],
            "5": ["3", "6", "7", "8"],
            "6": ["5"],
            "7": ["5", "8"],
            "8": ["5", "7", "9"],
            "9": ["8"],
        }
        start = "0"
        title = Text("Depth-First Search in a graph", font="Cascadia Code").to_edge(UP)
        self.play(Create(title))
        self.dfs(graph, start)
        self.wait()


class Random(Scene):
    def construct(self):
        graph = {
            "0": [("1", 2), ("2", 4)],
            "1": [("0", 2), ("2", 1), ("3", 5)],
            "2": [("0", 4), ("1", 1)],
            "3": [("1", 5), ("4", 2)],
            "4": [("3", 2), ("5", 2), ("6", 1)],
            "5": [("4", 2), ("6", 6)],
            "6": [("4", 1), ("5", 6), ("7", 3)],
            "7": [("6", 3)],
        }

        nodes_and_positions = {
            "0": LEFT * 4,
            "1": LEFT * 2 + UP,
            "2": LEFT * 2 + DOWN,
            "3": ORIGIN,
            "4": RIGHT * 2,
            "5": RIGHT * 4 + DOWN,
            "6": RIGHT * 4 + UP,
            "7": RIGHT * 7 + UP,
        }
        mGraph = MGraph(graph, nodes_and_positions, style=MGraphStyle.GREEN).move_to(
            ORIGIN
        )

        self.play(Create(mGraph))
        self.play(mGraph.animate.add_node("8", position=DOWN * 2), run_time=5)
        self.play(mGraph.animate.add_edge("3", "8", 1))
        self.play(mGraph.animate.shift(UP))
        mGraph["0"].set_highlight(RED, 8.0)
        self.play(mGraph["0"].animate.highlight())
        self.play(mGraph[("0", "1")].animate.highlight())
        self.play(mGraph["1"].animate.highlight())
        self.wait()


class ShowBackwardsEdge(Scene):
    def construct(self):
        graph = {"0": [("1", 4), ("2", 3)], "1": [], "2": []}

        nodes_and_positions = {
            "0": LEFT * 2 + UP * 2,
            "1": LEFT * 2 + DOWN * 2,
            "2": RIGHT * 2 + UP * 2,
        }

        mGraph = MGraph(graph, nodes_and_positions, style=MGraphStyle.GREEN)

        self.play(Create(mGraph))
        self.play(mGraph.animate.show_backward_edge("0", "2", 3, 0))
        self.wait()


"""
class NodeLayout(Scene):
    def construct(self):
        graph = {
            "0": ["1", "2"],
            "1": ["3", "4"],
            "2": ["5", "6"],
            "3": [],
            "4": [],
            "5": [],
            "6": [],
        }
        mGraph = MTree(graph, "0")
        self.play(Create(mGraph))
        self.wait()
"""


class Test(Scene):
    def construct(self):
        graph = {
            "0": ["1", "2"],
            "1": [],
            "2": [],
        }
        nodes_and_pos = {
            "0": LEFT * 5,
            "1": RIGHT,
            "2": DOWN,
        }
        mGraph = MGraph(graph, nodes_and_pos)
        self.play(Create(mGraph))
        self.wait()


class ReadMe(Scene):
    def construct(self):
        graph = {
            "A": [("C", 11), ("D", 7)],
            "B": [("A", 5), ("C", 3)],
            "C": [("A", 11), ("B", 3)],
            "D": [("A", 7), ("C", 4)],
        }
        nodes_and_positions = {
            "A": LEFT * 1.5,
            "B": UP * 2,
            "C": RIGHT * 1.5,
            "D": DOWN * 2,
        }

        mArray = (
            MArray([1, 2, 3], style=MArrayStyle.BLUE)
            .add_indexes()
            .scale(0.9)
            .add_label(Text("Array", font="Cascadia Code"))
            .to_edge(LEFT, 1)
        )

        mStack = (
            MStack([3, 7, 98, 1], style=MStackStyle.GREEN)
            .scale(0.8)
            .add_label(Text("Stack", font="Cascadia Code"))
            .move_to(ORIGIN)
        )

        mGraph = (
            MGraph(graph, nodes_and_positions, MGraphStyle.PURPLE)
            .add_label(Text("Graph", font="Cascadia Code"))
            .to_edge(RIGHT, 1)
        )

        self.play(Create(mArray))
        self.play(Create(mStack))
        self.play(Create(mGraph))
        self.wait()


class Logo(Scene):
    def construct(self):
        manim = Text("Manim", font="CMU Serif").scale(2).move_to(ORIGIN)
        # dsa = Text("DSA", font="CMU Serif").scale(2).next_to(manim, DOWN, DEFAULT_MOBJECT_TO_MOBJECT_BUFFER, RIGHT).shift(RIGHT*2)
        # obj = VGroup(manim, dsa).move_to(ORIGIN)
        style = MArrayStyle.PURPLE
        style.value["font"] = "Javiera"
        style.value["weight"] = BOLD
        mArray = (
            MArray(["D", "S", "A"], style=style)
            .add_indexes(DOWN)
            .shift(DOWN * 1.3 + RIGHT * 2.5)
        )
        self.play(Create(manim))
        self.play(Create(mArray))

        vertices = [1, 2, 3, 4]
        edges = [(1, 2), (2, 3), (3, 4), (1, 3), (1, 4)]
        g = Graph(
            vertices,
            edges,
            layout="planar",
            vertex_config={
                1: {"fill_color": ManimColor("#38f4af")},
                2: {"fill_color": ManimColor("#38f4af")},
                3: {"fill_color": ManimColor("#38f4af")},
                4: {"fill_color": ManimColor("#38f4af")},
            },
        )
        g[2].move_to([0, 0, 0])
        g[1].move_to([3, -0.1, 0])
        g[3].move_to([2.2, -0.4, 0])
        g[4].move_to([2.5, -0.8, 0])
        self.play(Create(g.shift([0.7, 0.37, 0])))
        self.play(VGroup(manim, mArray, g).animate.move_to(ORIGIN))
        self.wait()


class DfsRecursive(Scene):
    def dfs_helper(self, graph, mGraph, visited, prev, root):
        visited[root] = True
        self.play(mGraph[root].animate.highlight())
        for adj in graph[root]:
            if not visited[adj]:
                self.play(mGraph[(root, adj)].animate.highlight())
                self.dfs_helper(graph, mGraph, visited, prev, adj)
                self.play(mGraph[(root, adj)].animate.unhighlight())
        self.play(mGraph[root].animate.unhighlight())

    def dfs(self, graph, mGraph):
        visited = {}

        for node in graph:
            visited[node] = False

        for node in graph:
            if not visited[node]:
                self.dfs_helper(graph, mGraph, visited, None, node)

    def construct(self):
        graph = {
            "0": ["1", "2"],
            "1": ["0", "2", "3", "4"],
            "2": ["0", "1"],
            "3": ["1", "5"],
            "4": ["1"],
            "5": ["3", "6", "7", "8"],
            "6": ["5"],
            "7": ["5", "8"],
            "8": ["5", "7", "9"],
            "9": ["8"],
        }

        nodes_and_positions = {
            "0": LEFT * 6,
            "1": LEFT * 4 + UP,
            "2": LEFT * 4 + DOWN,
            "3": LEFT * 2,
            "4": LEFT * 2 + UP * 2,
            "5": ORIGIN,
            "6": LEFT * 2 + DOWN * 2,
            "7": RIGHT * 2 + DOWN * 2,
            "8": RIGHT * 2 + UP * 2,
            "9": RIGHT * 4 + UP * 2,
        }

        mGraph = (
            MGraph(graph, nodes_and_positions, style=MGraphStyle.BLUE)
            .move_to(ORIGIN)
            .shift(DOWN / 2)
        )

        title = Text("Depth-First Search in a graph", font="Cascadia Code").to_edge(UP)

        self.play(Create(title))
        self.play(Create(mGraph))

        self.dfs(graph, mGraph)
        self.wait()


class Kruskal(Scene):
    def find(self, parent, i):
        if parent[i] == i:
            return i
        return self.find(parent, parent[i])

    def union(self, parent, rank, x, y):
        xroot = self.find(parent, x)
        yroot = self.find(parent, y)
        if rank[xroot] < rank[yroot]:
            parent[xroot] = yroot
        elif rank[xroot] > rank[yroot]:
            parent[yroot] = xroot
        else:
            parent[yroot] = xroot
            rank[xroot] += 1

    def kruskal(self, graph, nodes_and_positions):
        mGraph = MGraph(graph, nodes_and_positions, style=MGraphStyle.PURPLE).move_to(
            ORIGIN
        )
        self.play(Create(mGraph))

        edges = []
        for u in graph:
            for v, weight in graph[u]:
                if (weight, u, v) not in edges and (weight, v, u) not in edges:
                    edges.append((weight, u, v))
        edges.sort()

        parent = {}
        rank = {}

        for node in graph:
            parent[node] = node
            rank[node] = 0

        mst_weight = 0

        for edge in edges:
            wt, u, v = edge
            x = self.find(parent, u)
            y = self.find(parent, v)
            if x != y:
                self.play(mGraph[(u, v)].animate.highlight(GREEN, 12))
                mst_weight += wt
                self.union(parent, rank, x, y)
            else:
                self.play(mGraph[(u, v)].animate.highlight(RED, 12))

        return mst_weight

    def construct(self):
        graph = {
            "0": [("1", 2), ("2", 4)],
            "1": [("0", 2), ("2", 1), ("3", 5), ("4", 5)],
            "2": [("0", 4), ("1", 1)],
            "3": [("1", 5), ("5", 2)],
            "4": [("1", 5)],
            "5": [("3", 2), ("6", 7), ("7", 2), ("8", 1)],
            "6": [("5", 7)],
            "7": [("5", 2), ("8", 6)],
            "8": [("5", 1), ("7", 6), ("9", 3)],
            "9": [("8", 3)],
        }

        nodes_and_positions = {
            "0": LEFT * 6,
            "1": LEFT * 4 + UP * 2,
            "2": LEFT * 4 + DOWN * 2,
            "3": LEFT * 2,
            "4": LEFT * 2 + UP * 2,
            "5": ORIGIN + RIGHT,
            "6": LEFT + DOWN * 2,
            "7": RIGHT * 3 + DOWN * 2,
            "8": RIGHT * 3 + UP * 2,
            "9": RIGHT * 5 + UP * 2,
        }

        title = (
            Text("Kruskal’s Algorithm for Minimum Spanning Tree", font="Cascadia Code")
            .scale(0.7)
            .to_edge(UP)
        )
        self.play(Create(title))
        total_weight = self.kruskal(graph, nodes_and_positions)
        text = Text("Total: " + str(total_weight), font="Cascadia Code").to_edge(DOWN)
        self.play(Create(text))
        self.wait()


class Prim(Scene):
    def prim(self, graph, nodes_and_positions, start):
        pq = []
        visited = {}
        mGraph = MGraph(graph, nodes_and_positions, style=MGraphStyle.PURPLE).move_to(
            ORIGIN
        )
        self.play(Create(mGraph))

        for node in graph:
            visited[node] = False

        res = 0

        heapq.heappush(pq, (0, None, start))

        while pq:
            wt, prev_node, u = heapq.heappop(pq)
            if visited[u]:
                self.play(mGraph[(prev_node, u)].animate.highlight(RED))
                continue

            visited[u] = True
            res += wt

            if prev_node is not None:
                self.play(mGraph[(prev_node, u)].animate.highlight(GREEN))

            self.play(mGraph[u].animate.highlight(GREEN))

            for adj in graph[u]:
                v, weight = adj
                if not visited[v]:
                    heapq.heappush(pq, (weight, u, v))
                    self.play(mGraph[(u, v)].animate.highlight(BLUE))

        return res

    def construct(self):
        graph = {
            "0": [("1", 2), ("2", 4)],
            "1": [("0", 2), ("2", 1), ("3", 5), ("4", 5)],
            "2": [("0", 4), ("1", 1)],
            "3": [("1", 5), ("5", 2)],
            "4": [("1", 5)],
            "5": [("3", 2), ("6", 7), ("7", 2), ("8", 1)],
            "6": [("5", 7)],
            "7": [("5", 2), ("8", 6)],
            "8": [("5", 1), ("7", 6), ("9", 3)],
            "9": [("8", 3)],
        }

        nodes_and_positions = {
            "0": LEFT * 6,
            "1": LEFT * 4 + UP,
            "2": LEFT * 4 + DOWN,
            "3": LEFT * 2,
            "4": LEFT * 2 + UP * 2,
            "5": ORIGIN,
            "6": LEFT * 2 + DOWN * 2,
            "7": RIGHT * 2 + DOWN * 2,
            "8": RIGHT * 2 + UP * 2,
            "9": RIGHT * 4 + UP * 2,
        }

        title = (
            Text("Prim's Algorithm for Minimum Spanning Tree", font="Cascadia Code")
            .scale(0.7)
            .to_edge(UP)
        )
        self.play(Create(title))
        total_weight = self.prim(graph, nodes_and_positions, "0")
        text = Text("Total: " + str(total_weight), font="Cascadia Code").to_edge(DOWN)
        self.play(Create(text))
        self.wait()



from __future__ import annotations

from manim import *


class MGraphStyle:
    class _DefaultStyle:
        def __init__(self):
            self.node_circle: dict = {"color": WHITE, "stroke_width": 6, "radius": 0.33}
            self.node_label: dict = {
                "color": WHITE,
                "font": "Cascadia Code",
                "font_size": 32,
                "weight": BOLD,
            }
            self.edge_line: dict = {
                "color": GRAY,
                "stroke_width": 7,
            }
            self.edge_tip: dict = {
                "stroke_width": 0.5,
                "fill_opacity": 1,
                "color": GRAY,
            }
            self.edge_weight: dict = {
                "color": WHITE,
                "font_size": 21,
                "font": "Javiera",
                "weight": BOLD,
            }
            self.start_distance: float = 0.2

    class _BlueStyle(_DefaultStyle):
        def __init__(self):
            super().__init__()
            self.node_circle: dict = {
                "color": BLUE_B,
                "fill_color": BLUE_D,
                "stroke_width": 6,
                "fill_opacity": 0.75,
                "radius": 0.33,
            }

    class _PurpleStyle(_DefaultStyle):
        def __init__(self):
            super().__init__()
            self.node_circle: dict = {
                "color": ManimColor("#eb97fc"),
                "fill_color": ManimColor("#8c46d6"),
                "stroke_width": 6,
                "fill_opacity": 0.75,
                "radius": 0.33,
            }

    class _GreenStyle(_DefaultStyle):
        def __init__(self):
            super().__init__()
            self.node_circle: dict = {
                "color": ManimColor("#b2ff8c"),
                "fill_color": ManimColor("#2ea556"),
                "stroke_width": 6,
                "fill_opacity": 0.75,
                "radius": 0.33,
            }

    DEFAULT = _DefaultStyle()
    BLUE = _BlueStyle()
    PURPLE = _PurpleStyle()
    GREEN = _GreenStyle()


class CollectionStyle:
    class _DefaultStyle:
        def __init__(self):
            self.square: dict = {
                "color": WHITE,
                "stroke_width": 6,
                "width": 1,
                "height": 1,
            }
            self.value: dict = {
                "color": WHITE,
                "font": "Cascadia Code",
                "font_size": 48,
                "weight": BOLD,
            }

    class _BlueStyle(_DefaultStyle):
        def __init__(self):
            super().__init__()
            self.square: dict = {
                "color": BLUE_B,
                "fill_color": BLUE_D,
                "stroke_width": 6,
                "fill_opacity": 1,
                "width": 1,
                "height": 1,
            }

    class _PurpleStyle(_DefaultStyle):
        def __init__(self):
            super().__init__()
            self.square: dict = {
                "color": ManimColor("#eb97fc"),
                "fill_color": ManimColor("#8c46d6"),
                "fill_opacity": 1,
                "stroke_width": 6,
                "width": 1,
                "height": 1,
            }

    class _GreenStyle(_DefaultStyle):
        def __init__(self):
            super().__init__()
            self.square: dict = {
                "color": ManimColor("#b2ff8c"),
                "fill_color": ManimColor("#2ea556"),
                "fill_opacity": 1,
                "stroke_width": 6,
                "width": 1,
                "height": 1,
            }

    DEFAULT = _DefaultStyle()
    BLUE = _BlueStyle()
    PURPLE = _PurpleStyle()
    GREEN = _GreenStyle()


class MStackStyle(CollectionStyle):
    class _DefaultStyle(CollectionStyle._DefaultStyle):
        def __init__(self):
            super().__init__()
            self.container: dict = {"color": RED}

    class _BlueStyle(_DefaultStyle, CollectionStyle._BlueStyle):
        def __init__(self):
            super().__init__()

    class _PurpleStyle(_DefaultStyle, CollectionStyle._PurpleStyle):
        def __init__(self):
            super().__init__()

    class _GreenStyle(_DefaultStyle, CollectionStyle._GreenStyle):
        def __init__(self):
            super().__init__()

    DEFAULT = _DefaultStyle()
    BLUE = _BlueStyle()
    PURPLE = _PurpleStyle()
    GREEN = _GreenStyle()


class MArrayStyle(CollectionStyle):
    class _DefaultStyle(CollectionStyle._DefaultStyle):
        def __init__(self):
            super().__init__()
            self.index: dict = {
                "color": WHITE,
                "font": "Cascadia Code",
                "font_size": 31,
            }

    class _BlueStyle(_DefaultStyle, CollectionStyle._BlueStyle):
        def __init__(self):
            super().__init__()
            self.index: dict = {
                "color": BLUE_D,
                "font": "Cascadia Code",
                "font_size": 31,
            }

    class _PurpleStyle(_DefaultStyle, CollectionStyle._PurpleStyle):
        def __init__(self):
            super().__init__()
            self.index: dict = {
                "color": ManimColor("#fabcff"),
                "font": "Cascadia Code",
                "font_size": 31,
            }

    class _GreenStyle(_DefaultStyle, CollectionStyle._GreenStyle):
        def __init__(self):
            super().__init__()
            self.index: dict = {
                "color": WHITE,
                "font": "Cascadia Code",
                "font_size": 31,
            }

    DEFAULT = _DefaultStyle()
    BLUE = _BlueStyle()
    PURPLE = _PurpleStyle()
    GREEN = _GreenStyle()


class VariableStyle(CollectionStyle):
    pass


# -----------Label configs-----------
DEFAULT_LABEL_ARGS: dict = {"color": BLUE_A, "font": "Cascadia Code", "font_size": 40}



from __future__ import annotations

from manim import *
from manim.typing import Vector3D

from manim_dsa.constants import *
from manim_dsa.m_collection.m_collection import MElement
from manim_dsa.utils.utils import *


class MVariable(MElement, Labelable):
    def __init__(
        self,
        value: str,
        style=VariableStyle.DEFAULT,
    ):
        self.style = style
        super().__init__(
            Rectangle(**self.style.square), Text(str(value), **self.style.value)
        )

    def add_label(
        self,
        text: Text,
        direction: Vector3D = UP,
        buff: float = 0.5,
        **kwargs,
    ):
        super().add_label(text, direction, buff, **kwargs)
        self += self.label
        return self



from manim import *

from manim_dsa.m_collection.m_stack import *


class Random(Scene):
    def construct(self):
        stack = MStack([1], style=MStackStyle.BLUE)
        self.play(Create(stack))
        self.play(stack.animate.append(8))
        self.play(stack.animate.pop())
        self.play(stack.animate.pop())
        self.play(stack.animate.scale(5))
        self.wait()



from __future__ import annotations

from typing import Any, override

from manim import *
from manim.typing import Point3D, Vector3D

from manim_dsa.constants import *
from manim_dsa.m_collection.m_collection import *
from manim_dsa.utils.utils import *


class MStack(MCollection):
    """Manim Stack: a class for visualizing the stack structure using the Manim animation engine.

    Parameters
    ----------
    arr : list, optional
        The initial list of values to populate the stack. Default is an empty list.
    buff : float, optional
        The buffer (margin) between elements in the stack. Default is 0.1.
    style : MStackStyle._DefaultStyle, optional
        The style configuration for the stack elements. Default is MStackStyle.DEFAULT.
    """

    def __init__(
        self,
        arr: list = [],
        buff: float = 0.1,
        style: MStackStyle._DefaultStyle = MStackStyle.DEFAULT,
    ):
        super().__init__(arr, UP, buff, style)

        elem = self.elements[0].square if self.elements else self._hidden_element.square
        container_height = (
            (len(arr) + 3) * elem.height
            if arr
            else self._hidden_element.square.height * 7
        )

        self.bottom_line: Line = Line(ORIGIN, [elem.width + 2 * buff, 0, 0]).next_to(
            elem, DOWN, buff
        )
        self.left_line: Line = Line([0, container_height, 0], ORIGIN).next_to(
            self.bottom_line, UL, 0
        )
        self.right_line: Line = self.left_line.copy().next_to(self.bottom_line, UR, 0)

        self.container: VGroup = VGroup(
            self.left_line, self.bottom_line, self.right_line
        )
        self += self.container
        self.move_to(ORIGIN)
        self.spawnpoint: Point3D = None

        # When the stack is scaled or moved,
        # the spawn_point of the objects must be changed as well
        def update_stack_attr(obj):
            obj.spawnpoint = obj.get_spawn_point()
            obj.margin = buff * self._hidden_element.square.width

        self.add_updater(update_stack_attr)

    def get_spawn_point(self) -> Point3D:
        """Calculates the drop point for new elements in the stack.

        Returns
        -------
        Point3D
            The spawn point position in 3D space.
        """
        return (
            self.bottom_line.get_center()
            + (UP * self.right_line.height)
            + UP * self._hidden_element.square.width
        )

    def append(self, value: Any) -> Self:
        """Appends a new value to the top of the stack.

        Parameters
        ----------
        value : Any
            The value to be added to the stack. It will be converted to a string representation.

        Returns
        -------
        self
            The instance of the :class:`MStack` with the newly appended element.
        """
        return super().append(value)

    @override_animate(append)
    def _append_animation(self, value: Any, anim_args=None) -> Succession:
        """Creates an animation for appending a new value to the stack.

        Parameters
        ----------
        value : Any
            The value to be added to the stack. It will be converted to a string representation.
        anim_args : dict, optional
            Additional animation arguments.

        Returns
        -------
        Succession
            The animation object representing the append operation.
        """
        self.append(value)
        new_pos = self.elements[-1].get_center()
        self.elements[-1].move_to(self.spawnpoint)

        return Succession(
            Create(self.elements[-1]),
            ApplyMethod(self.elements[-1].move_to, new_pos),
            **anim_args,
            group=self,
        )

    @override
    def pop(self) -> Self:
        """Removes the top element from the stack.

        Returns
        -------
        self
            The instance of the :class:`MStack` with the top element removed.
        """
        return super().pop(len(self.elements) - 1)

    @override_animate(pop)
    def _pop_animation(self, anim_args=None) -> Succession:
        """Creates an animation for removing the top element from the stack.

        Parameters
        ----------
        anim_args : dict, optional
            Additional animation arguments.

        Returns
        -------
        Succession
            The animation object representing the pop operation.
        """
        popped_element = self.elements[-1].copy()
        self.pop()
        return Succession(
            ApplyMethod(popped_element.move_to, self.spawnpoint),
            FadeOut(popped_element),
            **anim_args,
            group=VGroup(self, popped_element),
        )

    def add_label(
        self,
        text: Text,
        direction: Vector3D = UP,
        buff: float = 0.5,
        **kwargs,
    ) -> Self:
        """Adds a label to the stack.

        Parameters
        ----------
        text : Text
            The label text.
        direction : Vector3D, optional
            The direction in which to position the label. Default is UP.
        buff : float, optional
            The distance (buffer) between the stack and the label. Default is 0.5.
        **kwargs :
            Additional keyword arguments that are passed to the function next_to() of the
            underlying add_label method.

        Returns
        -------
        self
            The instance of the :class:`MStack` with the label added.
        """
        super().add_label(text, direction, buff, **kwargs)
        self.label.move_to(self.get_spawn_point())
        return self



from manim import *

from manim_dsa.m_collection.m_array import *


class Init(Scene):
    def construct(self):
        mArray = (
            MArray([1, 2, 3, 4, 5], style=MArrayStyle.BLUE)
            .add_indexes()
            .scale(0.9)
            .add_label(Text("Array", font="Cascadia Code"))
        )
        self.play(Create(mArray))
        self.wait()


class RandomOperations(Scene):
    def construct(self):
        arr = [1, 2, 3]
        mArray = (
            MArray(arr, style=MArrayStyle.PURPLE)
            .add_indexes(DOWN)
            .add_label(Text("Array", **DEFAULT_LABEL_ARGS), DOWN)
        )
        mArray.shift(UP * 1.5 + LEFT * 4)
        self.play(Create(mArray))
        self.play(mArray.animate.append(4))
        self.play(mArray.animate.append(5))
        self.play(mArray.animate.scale(0.5))
        self.play(mArray.animate.append(6))
        self.play(mArray.animate.append(7))
        self.play(mArray.animate.pop(0))
        self.play(mArray.animate.shift(RIGHT))
        self.play(mArray.animate.pop(2))
        self.play(mArray[0].value.animate.set_fill(RED))
        self.play(mArray[0].index.animate.set_fill(RED))
        self.play(mArray.animate.swap(0, 3, path_arc=PI / 2))
        self.play(mArray[0].value.animate.set_fill(GREEN))
        self.play(mArray[0].index.animate.set_fill(GREEN))
        self.wait(1)



from __future__ import annotations

from manim import *
from manim.typing import Vector3D


def set_text(old_manim_text: Text, new_text: str) -> Text:
    """
    Replace the content of an existing Manim Text object with new text while preserving its style.

    Parameters
    ----------
    old_manim_text : Text
        The original Manim Text object whose content is being replaced.
    new_text : str
        The new text content to set.

    Returns
    -------
    Text
        A new Text object with the updated content, matching the style and position of the original text.
    """
    NewText = type(old_manim_text)
    res = (
        NewText(
            str(new_text), font=old_manim_text.font, font_size=old_manim_text.font_size
        )
        .match_style(old_manim_text)
        .move_to(old_manim_text)
    )
    return res


def TextReplace(scene: Scene, scene_mobj1: Group, mObj1: Text, mObj2: Text):
    """
    Replace one text object with another in a scene using animations.

    Parameters
    ----------
    scene : Scene
        The Manim Scene where the animation takes place.
    scene_mobj1 : Group
        The group containing `mObj1`, which is replaced in this group.
    mObj1 : Text
        The original text object to be replaced.
    mObj2 : Text
        The text object whose content will replace `mObj1`.

    Notes
    -----
    The animation consists of fading out the old text (`mObj1`) while animating
    the new text (`mObj2`) into its position.
    """
    old_mobj = mObj1.copy()
    scene_mobj1 -= mObj1
    mObj1 = set_text(mObj1, str(mObj2.text))
    scene_mobj1 += mObj1
    new_mobj = mObj1.copy()
    mObj1.set_opacity(0)
    scene.play(
        ReplacementTransform(mObj2.copy(), new_mobj),
        ApplyMethod(old_mobj.set_opacity, 0),
    )
    mObj1.set_opacity(1)
    new_mobj.set_opacity(0)


class Labelable:
    """
    A mixin class that provides functionality to add a label to Manim objects.

    Attributes
    ----------
    label : Text or None
        The label associated with the object, if any.
    """

    def __init__(self):
        """
        Initialize the Labelable object with no label.
        """
        super().__init__()
        self.label = None

    def add_label(
        self,
        text: Text,
        direction: Vector3D = UP,
        buff: float = 0.5,
        **kwargs,
    ) -> Labelable:
        """
        Add a label to the object.

        Parameters
        ----------
        text : Text
            The Text object to use as the label.
        direction : Vector3D, optional
            The direction to place the label relative to the object (default is `UP`).
        buff : float, optional
            The distance between the object and the label (default is 0.5).
        **kwargs : dict
            Additional keyword arguments for positioning.

        Returns
        -------
        Labelable
            The instance with the label added.
        """
        self.label = text
        self.label.next_to(self, direction, buff, **kwargs)
        return self

    def has_label(self) -> bool:
        """
        Check if the object has a label.

        Returns
        -------
        bool
            `True` if the object has a label, otherwise `False`.
        """
        return self.label is not None


class Highlightable:
    """
    A mixin class that provides functionality to highlight and unhighlight Manim objects.

    Attributes
    ----------
    highlighting : VMobject or None
        The highlight effect associated with the object, if any.
    """

    def __init__(self):
        """
        Initialize the Highlightable object with no highlighting.
        """
        super().__init__()
        self.__target = None
        self.highlighting = None

    def _add_highlight(self, target: VMobject):
        """
        Internal method to set up highlighting for a target VMobject.

        Parameters
        ----------
        target : VMobject
            The object to highlight.
        """
        self.__target = target
        self.highlighting = (
            target.copy().set_fill(opacity=0).set_z_index(self.__target.z_index + 1)
        )
        self.set_highlight()

    def highlight(
        self, stroke_color: ManimColor = RED, stroke_width: float = 8
    ) -> Highlightable:
        """
        Highlight the object with the specified stroke color and width.

        Parameters
        ----------
        stroke_color : ManimColor, optional
            The color of the highlight stroke (default is `RED`).
        stroke_width : float, optional
            The width of the highlight stroke (default is 8).

        Returns
        -------
        Highlightable
            The instance with the highlight applied.
        """
        self.set_highlight(stroke_color, stroke_width)
        # Since the target object could have been scaled or moved, scale and move self.highlighting
        self.highlighting.width = self.__target.width
        self.highlighting.height = self.__target.height
        self.highlighting.move_to(self.__target)
        self += self.highlighting
        return self

    @override_animate(highlight)
    def _highlight_animation(
        self,
        stroke_color: ManimColor = RED,
        stroke_width: float = 8,
        anim_args=None,
    ) -> Animation:
        """
        Animation for highlighting the object.

        Parameters
        ----------
        stroke_color : ManimColor, optional
            The color of the highlight stroke (default is `RED`).
        stroke_width : float, optional
            The width of the highlight stroke (default is 8).
        anim_args : dict, optional
            Additional arguments for the animation.

        Returns
        -------
        Animation
            The animation for highlighting.
        """
        self.highlight(stroke_color, stroke_width)
        return Create(self.highlighting, **anim_args)

    def set_highlight(self, stroke_color: ManimColor = RED, stroke_width: float = 8):
        """
        Set the highlight properties.

        Parameters
        ----------
        stroke_color : ManimColor, optional
            The color of the highlight stroke (default is `RED`).
        stroke_width : float, optional
            The width of the highlight stroke (default is 8).
        """
        self.highlighting.set_stroke(stroke_color, stroke_width)

    def unhighlight(self) -> Highlightable:
        """
        Remove the highlight from the object.

        Returns
        -------
        Highlightable
            The instance with the highlight removed.
        """
        self -= self.highlighting
        return self

    @override_animate(unhighlight)
    def _unhighlight_animation(self, anim_args=None) -> Animation:
        """
        Animation for unhighlighting the object.

        Parameters
        ----------
        anim_args : dict, optional
            Additional arguments for the animation.

        Returns
        -------
        Animation
            The animation for unhighlighting.
        """
        if anim_args is None:
            anim_args = {}

        self.unhighlight()
        return FadeOut(self.highlighting, **anim_args)



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import os
import sys
from datetime import datetime

from manim.utils.docbuild.module_parsing import parse_module_attributes

import manim_dsa

# -- Path setup --------------------------------------------------------------
# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

sys.path.insert(0, os.path.abspath("."))

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "manim-dsa"
copyright = f"{datetime.now().year}, Fabio Missagia"
author = "Fabio Missagia"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.duration",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.intersphinx",
    "sphinx.ext.viewcode",
    "manim.utils.docbuild.manim_directive",
    "sphinx.ext.napoleon",
]

templates_path = ["_templates"]
exclude_patterns = []

# displays shorter function names that are documented via autodoc - sphinx.ext.autosummary
add_module_names = False

# displays type hint defaults - sphinx_autodoc_typehints
typehints_defaults = "comma"

autodoc_default_options = {
    "members": True,  # Mostra i membri della classe
}

# generate documentation from type hints
ALIAS_DOCS_DICT = parse_module_attributes()[0]
autodoc_typehints = "description"
autodoc_type_aliases = {
    alias_name: f"~manim.{module}.{alias_name}"
    for module, module_dict in ALIAS_DOCS_DICT.items()
    for category_dict in module_dict.values()
    for alias_name in category_dict
}

# allows external documentation to be referred - sphinx.ext.intersphinx
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "manim": ("https://docs.manim.community/en/stable/", None),
}

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "furo"
# html_favicon = str(Path("_static/favicon.ico"))
html_static_path = ["_static"]
html_title = f"Manim DSA v{manim_dsa.__version__}"
html_css_files = ["custom.css"]

html_theme_options = {
    "source_repository": "https://github.com/F4bbi/manim-dsa/",
    "source_branch": "main",
    "source_directory": "docs/source/",
    "top_of_page_button": None,
    "light_logo": "light-logo.svg",
    "dark_logo": "dark-logo.svg",
}



# This script is used to remove the <pre> tags from the HTML files in the build directory that are generated by Manim.
# For more information, please refer to the following link: https://github.com/ManimCommunity/manim/issues/3994

import argparse
import os

from bs4 import BeautifulSoup

parser = argparse.ArgumentParser(description="Process a file path.")
parser.add_argument("path", type=str, help="The path of html build.")
args = parser.parse_args()


def clean_build():
    def remove_pre_tags(html_content):
        soup = BeautifulSoup(html_content, "html.parser")
        for tag in soup.find_all("pre", {"data-manim-binder": True}):
            tag.decompose()
        return str(soup)

    for root, _, files in os.walk(args.path):
        for filename in files:
            if filename.endswith(".html"):
                file_path = os.path.join(root, filename)
                with open(file_path, encoding="utf-8") as file:
                    html_content = file.read()
                cleaned_html = remove_pre_tags(html_content)
                with open(file_path, "w", encoding="utf-8") as file:
                    file.write(cleaned_html)


clean_build()



from __future__ import annotations

from typing import Any, override

from manim import *
from manim.typing import Vector3D

from manim_dsa.constants import *
from manim_dsa.m_collection.m_collection import *
from manim_dsa.utils.utils import *


class MIndexedElement(MElement):
    """An extension of the MElement class that includes an index for each element.

    Parameters
    ----------
    square : Rectangle
        The rectangle representing the visual boundary of the element.
    value : Text
        The text representing the value contained in the element.
    """

    def __init__(self, square: Rectangle, value: Text):
        super().__init__(square, value)

    def set_index(self, new_index: Any) -> Self:
        """
        Updates the index of the element to a new value.

        Parameters
        ----------
        new_index : Any
            The new index value to set. It will be converted to a string representation.

        Returns
        -------
        self
            The instance of the :class:`MIndexedElement` with the updated index.
        """
        self -= self.index
        self.index = utils.set_text(self.index, str(new_index))
        self += self.index
        return self

    def add_index(
        self,
        index: Text,
        direction: Vector3D = UP,
        buff: float = DEFAULT_MOBJECT_TO_MOBJECT_BUFFER,
    ) -> Self:
        """
        Adds an index to the element, positioning it relative to the element.

        Parameters
        ----------
        index : Text
            The text object representing the index.
        direction : Vector3D, optional
            The direction in which to position the index relative to the element. Default is UP.
        buff : float, optional
            The distance (buffer) between the element and the index.
            Default is DEFAULT_MOBJECT_TO_MOBJECT_BUFFER.

        Returns
        -------
        self
            The instance of the :class:`MIndexedElement` with the updated index.
        """
        self.index = index.next_to(self.square, direction, buff)
        self += self.index
        return self


class MArray(MCollection):
    """Manim Array: a class for visualizing the array structure using the Manim animation engine.

    Parameters
    ----------
    arr : list, optional
        The initial list of values to populate the array. Default is an empty list.
    direction : Vector3D, optional
        The direction in which to arrange the elements. Default is RIGHT.
    margin : float, optional
        The margin between elements in the array. Default is 0.
    style : MArrayStyle._DefaultStyle, optional
        The style configuration for the elements. Default is MArrayStyle.DEFAULT.
    """

    def __init__(
        self,
        arr: list = [],
        direction: Vector3D = RIGHT,
        style: MArrayStyle._DefaultStyle = MArrayStyle.DEFAULT,
    ):
        self._index_enabled: bool = False
        self._index_dir = None

        super().__init__(arr, direction, 0, style)

        self._hidden_element = MIndexedElement(
            self._hidden_element.square,
            self._hidden_element.value,
        )
        self += self._hidden_element

    def append(self, value: Any) -> Self:
        """Appends a new element to the end of the array. If indexing is enabled,
        the new element will also be assigned an index based on its position in the array.

        Parameters
        ----------
        value : Any
            The value to append. It will be converted to a string representation.

        Returns
        -------
        self
            The instance of the :class:`MArray` with the newly appended element.
        """
        self._update_style()

        new_elem = MIndexedElement(
            Rectangle(**self.style.square), Text(str(value), **self.style.value)
        )

        self._append_helper(new_elem)

        if self._index_enabled:
            index = Text(str(len(self.elements) - 1), **self.style.index)
            self.elements[-1].add_index(index, self._index_dir, self._get_index_buff())

        return self

    @override_animate(append)
    def _append_animation(self, value: Any, anim_args=None) -> Animation:
        """Animates the addition of a new element to the array.

        Parameters
        ----------
        value : Any
            The value to append. It will be converted to a string representation.
        anim_args : dict, optional
            Additional arguments for the animation. Default is None.

        Returns
        -------
        Animation
            An animation that displays the new element being written into the array.
        """
        return super()._append_animation(value, anim_args)

    @override
    def _update_style(self) -> None:
        super()._update_style()
        if self._index_enabled:
            self.style.index["font_size"] = self._hidden_element.index.font_size

    def pop(self, index: int = -1) -> Self:
        """Removes the element at the specified index and shifts all subsequent elements accordingly.
        If indexing is enabled, it also updates the indices of the remaining elements.

        Parameters
        ----------
        index : int, optional
            The index of the element to be removed. Default is -1, which removes the last element.

        Returns
        -------
        self
            The instance of the :class:`MArray` with the specified element removed.
        """
        if len(self.elements):
            popped_element = self.elements[index].copy()
            super().pop(index)

            if self._index_enabled:
                self._set_index_from(
                    index, len(self.elements) - 1, popped_element.index
                )
        return self

    @override_animate(pop)
    def _pop_animation(self, index: int = -1, anim_args=None) -> Animation:
        """Animates the removal of an element from the array.

        Parameters
        ----------
        index : int, optional
            The index of the element to be removed. Default is -1, which removes the last element.
        anim_args : dict, optional
            Additional arguments for the animation. Default is None.

        Returns
        -------
        Animation
            An animation showing the element being removed and the indices being updated.
        """
        popped_element = self._logic_pop(index)

        elem_shift = VGroup(*self.elements[index:])

        anims = [
            FadeOut(popped_element),
            ApplyMethod(elem_shift.shift, -(self._dir * popped_element.square.width)),
        ]

        if self._index_enabled:
            anims.append(
                ApplyMethod(
                    self._set_index_from,
                    index,
                    len(self.elements) - 1,
                    popped_element.index.copy(),
                )
            )

        return Succession(*anims, **anim_args, group=VGroup(self, popped_element))

    def _set_index_from(self, start: int, end: int, popped_index: MElement) -> Self:
        old_index = popped_index
        for i in range(start, end + 1):
            curr = self.elements[i]
            new_index = old_index.copy().move_to(curr.index)
            old_index = curr.index
            curr -= curr.index
            curr.index = new_index
            curr += curr.index
        return self

    def _get_index_buff(self) -> float:
        square_center = self._hidden_element.square.get_center()
        index_center = self._hidden_element.index.get_center()
        offset = self._hidden_element.square.width / 2

        if np.array_equal(self._dir, UP) or np.array_equal(self._dir, DOWN):
            offset += self._hidden_element.index.width / 2
            direction_offset = [offset, 0, 0]
        else:
            offset += self._hidden_element.index.height / 2
            direction_offset = [0, offset, 0]

        if np.array_equal(self._index_dir, UP) or np.array_equal(
            self._index_dir, RIGHT
        ):
            return (index_center - square_center - direction_offset)[1]
        else:
            return (square_center - index_center - direction_offset)[1]

    def _visual_swap(self, i: int, j: int) -> None:
        elem_i = self.elements[i]
        elem_j = self.elements[j]
        elem_i_group = VGroup(elem_i.square, elem_i.value)
        temp = elem_i_group.copy()
        elem_j_group = VGroup(elem_j.square, elem_j.value)
        elem_i_group.move_to(elem_j_group, DOWN)
        elem_j_group.move_to(temp, DOWN)

    def _logic_swap(self, i: int, j: int) -> None:
        # Element swap
        # We have to remove first them from the scene to work correctly
        self -= self.elements[i]
        self -= self.elements[j]

        self.elements[i], self.elements[j] = self.elements[j], self.elements[i]

        if self._index_enabled:
            # Index swap
            # We have to remove first them from the scene to work correctly
            self.elements[i] -= self.elements[i].index
            self.elements[j] -= self.elements[j].index

            self.elements[i].index, self.elements[j].index = (
                self.elements[j].index,
                self.elements[i].index,
            )

            # We can add the indexes again
            self.elements[i] += self.elements[i].index
            self.elements[j] += self.elements[j].index

        # We can add the elements again
        self += self.elements[i]
        self += self.elements[j]

    def add_indexes(
        self,
        direction: Vector3D = UP,
        buff: float = DEFAULT_MOBJECT_TO_MOBJECT_BUFFER,
    ) -> Self:
        """Adds indexes to each element in the array, displaying them in the specified direction.

        Parameters
        ----------
        direction : Vector3D, optional
            The direction in which to display the indices relative to the elements.
            Default is UP.
        buff : float, optional
            The buffer distance between the element and its index.
            Default is DEFAULT_MOBJECT_TO_MOBJECT_BUFFER.

        Returns
        -------
        self
            The instance of the :class:`MArray` with the specified element removed.

        Raises
        ------
        Exception
            If the specified direction is parallel to the array's growth direction.

        Notes
        -----
        If indices are already enabled, this method returns immediately without making any changes.
        """
        if self._index_enabled:
            return self
        if np.array_equal(np.abs(self._dir), np.abs(direction)):
            raise Exception(
                "The direction given is parallel to array growth direction!"
            )

        self._index_enabled = True
        self._index_dir = direction

        self._hidden_element.add_index(
            Text("0", **self.style.index, fill_opacity=0), direction, buff
        )
        for i in range(len(self.elements)):
            self.elements[i].add_index(
                Text(str(i), **self.style.index), direction, buff
            )

        return self
